/* $Id: eval.y,v 1.1 1994/07/09 19:54:20 ryo Exp $
 *
 *	\[XR[hWFl[^
 *	xt@Cp[T
 *	Copyright (C) 1989,1990 K.Abe
 *	All rights reserved.
 *
 */

%{
#include	<stdio.h>
#include	<string.h>
#include	<stdlib.h>
#include	<ctype.h>
#include	"estruct.h"
#include	"global.h"
#include	"disasm.h"
#include	"label.h"
#include	"generate.h"	/* symset */
#include	"labelfile.h"
#include	"offset.h"	/* depend_address */
#include	"hex.h"		/* strend */
#include	"etc.h"
#include	"table.h"

/*
#define	YYDEBUG	1
*/
extern int	yydebug;

parse_mode	ParseMode;
char		*Lexptr;

int		Eval_Value;
address		Eval_TableTop;
opesize 	Eval_ID;
address		Eval_PC;
int		Eval_Bytes;
char		Eval_ResultString[256];
int		Eval_Break;

int	Eval_Count;

struct expbuf {
    boolean	isvalue;
    boolean	registed;
    address	value;
    char	*bufptr;
};

extern int	strncmpi( char* , char* , int );

static int		yylex( void );
static void		yyerror( char* );
static void		tabledesc( int , opesize );
static unsigned long	peek( address , opesize );
static unsigned long	extend( unsigned long , opesize );
static struct expbuf	*storestr( char* );
static struct expbuf	*storeexp( address , int );
static struct expbuf	*store( struct expbuf * );
static void		breakjob( unsigned long );
static void		itoxd_by_size( char* , ULONG , opesize );


#define	EXPSTACKSIZE	32
static struct expbuf	*ExpStack[ EXPSTACKSIZE ];
static int		ExpStackCounter = 0;

#define	PEEK_LONG(a)\
	(( *(a) << 24 ) + ( *( (a) + 1 ) << 16 ) + ( *( (a) + 2 ) << 8 ) + *( (a) + 3 ))
#define	PEEK_WORD(a)\
	(( *(a) << 8 ) + *( (a) + 1 ))


%}

/*

	Begin Grammer

*/
%union {
	unsigned long	val;
	char			*str;
	struct expbuf	*expbuf;
}
%token	<str>		_STRING
%token	<val>		SIZEID
%token	<val>		NUMBER
%token	<val>		PEEKB PEEKW PEEKL BREAK EXTW EXTL
%type	<expbuf>	str_or_exp	/* exprstring */
%type	<val>		sizeid
%type	<val>		tabledesc
%type	<val>		expr logical_AND_expr equality_expr relational_expr
%type	<val>		additive_expr mul_expr factor element
%left	'<' LE '>'	GE EQUEQU NOTEQU
%left	'(' ')' '{' '}'
%left	','
%left	OROR
%left	ANDAND
%left	'+' '-'
%left	'*' '/'
%left	'!'
%left	UNARYMINUS

%%

tabledesc : BREAK expr				{ breakjob( $2 ); }
	| sizeid				{ tabledesc(  1 , $1 ); }
	| sizeid '[' expr ']'			{ tabledesc( $3 , $1 ); }
	| sizeid exprstring			{ tabledesc(  1 , $1 ); }
	| sizeid '[' expr ']' exprstring	{ tabledesc( $3 , $1 ); }
	;

sizeid : SIZEID					{ $$ = $1; }
	| '@' SIZEID				{ $$ = $2; }
	;

exprstring : str_or_exp				{ store( $1 ); }
	| exprstring ',' str_or_exp 		{ store( $3 ); }
	;

str_or_exp : _STRING				{ $$ = storestr( $1 ); }
	| expr					{ $$ = storeexp( (address)$1 , 0 ); }
	| '{' expr '}'				{ $$ = storeexp( (address)$2 , 1 ); }
	;

expr : logical_AND_expr				{ $$ = $1; }
	| expr OROR logical_AND_expr		{ $$ = $1 || $3; }
	;

logical_AND_expr : equality_expr		{ $$ = $1; }
	| logical_AND_expr ANDAND equality_expr	{ $$ = $1 && $3; }
	;

equality_expr : relational_expr			{ $$ = $1; }
	| equality_expr EQUEQU relational_expr	{ $$ = ( $1 == $3 ); }
	| equality_expr NOTEQU relational_expr	{ $$ = ( $1 != $3 ); }
	;

relational_expr : additive_expr			{ $$ = $1; }
	| relational_expr '<' additive_expr	{ $$ = ( $1 <  $3 ); }
	| relational_expr LE additive_expr	{ $$ = ( $1 <= $3 ); }
	| relational_expr '>' additive_expr	{ $$ = ( $1 >  $3 ); }
	| relational_expr GE additive_expr	{ $$ = ( $1 >= $3 ); }
	;

additive_expr : mul_expr			{ $$ = $1; }
	| additive_expr '+' mul_expr		{ $$ = $1 + $3; }
	| additive_expr '-' mul_expr		{ $$ = $1 - $3; }
	;

mul_expr : factor				{ $$ = $1; }
	| mul_expr '*' factor			{ $$ = $1 * $3; }
	| mul_expr '/' factor			{ $$ = $1 / $3; }
	| mul_expr '%' factor			{ $$ = $1 % $3; }
	;

factor : '-' factor %prec UNARYMINUS	 	{ $$ = - $2; }
	| '!' factor				{ $$ = ! $2; }
	| element				{ $$ = $1; }
	;

element : '(' expr ')'				{ $$ = $2; }
	| NUMBER				{ $$ = $1; }
	| PEEKB '(' expr ')'	{ $$ = peek( (address)$3 , BYTESIZE ); }
	| PEEKW '(' expr ')'	{ $$ = peek( (address)$3 , WORDSIZE ); }
	| PEEKL '(' expr ')'	{ $$ = peek( (address)$3 , LONGSIZE ); }
	| EXTW '(' expr ')'	{ $$ = extend( (unsigned)$3 , WORDSIZE ); }
	| EXTL '(' expr ')'	{ $$ = extend( (unsigned)$3 , LONGSIZE ); }
	;
%%

/*

	End of Grammer

*/
static int	yylex()
{
    struct idpair {
	char	*idstr;
	int	idnum;
    } idbuf[] = {
	"dc.b" , BYTESIZE , "dc.w" , WORDSIZE	, "dc.l" , LONGSIZE ,
	"byte" , BYTEID	, "even" , EVENID , "cr" , CRID	,
	"asciz" , ASCIIZID , "asciiz" , ASCIIZID , "ascii" , ASCIIID ,
	"lascii" , LASCIIID ,
#if 0
	"long" , LONGID , "word" , "WORDID" ,
#endif
	NULL , 0 ,
    } , *idptr;
    
    int		c;
    
    while( ( c = *Lexptr++ ) == ' ' || c == '\t' || c == '\n' )
	;
    if( c == '\0' )
	return 0;
    
    if( isdigit(c) ) {
	yylval.val = c - '0';
	while( isdigit( c = *Lexptr++ ) )
	    yylval.val = yylval.val * 10 + c - '0';
	Lexptr--;
	return NUMBER;
    }
    if( c == '$' ) {
	yylval.val = 0;
	if( !isxdigit( *Lexptr ) )
	    return 1;
	while( isxdigit( c = *Lexptr++ ) )
	    yylval.val = yylval.val * 16 + ( toupper(c) >= 'A' ? toupper(c)-'A'+10 : c-'0' );
	Lexptr--;
	return NUMBER;
    }
    if( c == '"' ) {
	char	*term;

	if( ( term = strchr( Lexptr , '"' ) ) == NULL )
	    return 1;
	yylval.str = Malloc( term - Lexptr + 1 );
	strncpy( yylval.str , Lexptr , term - Lexptr );
	yylval.str[ term - Lexptr ] = '\0';
	Lexptr = term + 1;
	return _STRING;
    }
    if( c == '<' ) {
	if( *Lexptr == '=' ) {
	    Lexptr++;
	    return LE;
	}
	return c;
    }
    if( c == '>' ) {
	if( *Lexptr == '=' ) {
	    Lexptr++;
	    return GE;
	}
	return c;
    }
    if( c == '|' ) {
	if( *Lexptr == '|' ) {
	    Lexptr++;
	    return OROR;
	}
	return c;
    }
    if( c == '&' ) {
	if( *Lexptr == '&' ) {
	    Lexptr++;
	    return ANDAND;
	}
	return c;
    }
    if( c == '=' ) {
	if( *Lexptr == '=' ) {
	    Lexptr++;
	    return EQUEQU;
	}
	return c;
    }
    if( c == '!' ) {
	if( *Lexptr == '=' ) {
	    Lexptr++;
	    return NOTEQU;
	}
	return c;
    }
#define	strdup(s)	({ char *_s = alloca(strlen(s)+1) ; strcpy(_s,s);})
    if( iscsymf(c) ) {
	char	*ptr , *token = strdup( Lexptr - 1 );
	int		adj;
#undef strdup
	for( ptr = token ; iscsym( *ptr ) || *ptr == '.' ; ptr++ )
	;
	*ptr = '\0';	/* now tmp point token */
	adj = ptr - token - 1;

	for( idptr = idbuf ; idptr->idstr != NULL ; idptr++ ) {
	    if( strcmpi( token , idptr->idstr ) == 0 ) {
		yylval.val = (unsigned long)idptr->idnum;
		Lexptr += adj;
		return SIZEID;
	    }
	}
	if( strcmpi( token , "tabletop" ) == 0 ) {
	    yylval.val = (unsigned long)Eval_TableTop;
	    Lexptr += adj;
	    return NUMBER;
	}
	if( strcmpi( token , "pc" ) == 0 ) {
	    yylval.val = (unsigned long)Eval_PC;
	    Lexptr += adj;
	    return NUMBER;
	}
	if( strcmpi( token , "break" ) == 0 ) {
	    Lexptr += adj;
	    return BREAK;
	}
	if( strncmpi( token , "peek." , 5 ) == 0 ) {
	    Lexptr += adj;
	    switch( tolower( token[5] ) ) {
	    case 'b' :
		return PEEKB;
	    case 'w' :
		return PEEKW;
	    case 'l' :
		return PEEKL;
	    }
	}
	if( strncmpi( token , "ext." , 4 ) == 0 ) {
	    Lexptr += adj;
	    switch( tolower( token[4] ) ) {
	    case 'w' :
		return EXTW;
	    case 'l' :
		return EXTL;
	    }
	}
	free( token );
    }
    return c;
}


static void	yyerror( s )
char	*s;
{
    fprintf( stderr , "%s\n" , s );
}


static void	tabledesc( num , id )
int	num;
opesize	id;
{
    address	val;
    int		i;

    Eval_ID = id;
    if( Eval_Count == 0 ) {		/* easy trick ? */
	if( id == ASCIIID || id == BYTEID || id == WORDID || id == LONGID ) { 
	    Eval_Count = 1;
	    Eval_Bytes = num;
	} else
	    Eval_Count = num;
    }


    if( ExpStackCounter == 0 && ParseMode == PARSE_GENERATING ) {
	switch( id ) {
	case LONGSIZE :
	    if( depend_address( Eval_PC ) )
		make_proper_symbol( strend( Eval_ResultString ) ,
				    *(address*)( Eval_PC + Ofst ) );
	    else
		itoxd_by_size( strend( Eval_ResultString ) ,
			      PEEK_LONG( Eval_PC + Ofst ) , LONGSIZE );
	    break;
	case WORDSIZE :
	    itoxd_by_size( strend( Eval_ResultString ) ,
			  PEEK_WORD( Eval_PC + Ofst ) , WORDSIZE );
	    break;
	case BYTESIZE :
	    itoxd_by_size( strend( Eval_ResultString ) ,
			  *(UBYTE*)( Eval_PC + Ofst ) , BYTESIZE );
	    break;
	    default :	/* reduce warning message */
		;
	}
    }
    for( i = 0 ; i < ExpStackCounter ; i++ ) {
	switch( ParseMode ) {
	case PARSE_ANALYZING :
	    if( !( ExpStack[ i ]->isvalue ) )
		free( ExpStack[ i ]->bufptr );
	    free( ExpStack[ i ] );
	    break;
	case PARSE_GENERATING :
	    if( ExpStack[ i ]->isvalue ) {
		val = ExpStack[ i ]->value;
		if( ExpStack[ i ]->registed )
		    make_proper_symbol( strend( Eval_ResultString ) , val );
		else
		    itoxd_by_size( strend( Eval_ResultString ) , (ULONG)val , Eval_ID );
	    } else {
		strcat( Eval_ResultString , ExpStack[ i ]->bufptr );
		free( ExpStack[ i ]->bufptr );
	    }
	    free( ExpStack[ i ] );
	    break;
	}
    }
    ExpStackCounter = 0;
}


static struct expbuf	*storestr( s )
char	*s;
{
    struct expbuf	*expbufptr;

    expbufptr = Malloc( sizeof( struct expbuf ) );
    expbufptr->bufptr = s;
    expbufptr->isvalue = FALSE;
    return expbufptr;
}


static struct expbuf	*storeexp( v , mode )
address	v;
int	mode;
{
    struct expbuf	*expbufptr;
   
    expbufptr = Malloc( sizeof( struct expbuf ) );
    expbufptr->value = v;
    expbufptr->isvalue = TRUE;
    
    if( mode == 1 ) {
	expbufptr->registed = TRUE;
	if( ParseMode == PARSE_ANALYZING && !regist_label( v , DATLABEL | UNKNOWN ) )
	    printf( "??? address %x\n" , v );
    } else
	expbufptr->registed = FALSE;
    return expbufptr;
}


static struct expbuf	*store( s )
struct expbuf	*s;
{
    ExpStack[ ExpStackCounter++ ] = s;
    if( ExpStackCounter == EXPSTACKSIZE )
	err( 1 , "]X^bNӂ܂\n" );
    /*
      switch( s->isvalue ) {
      case TRUE :
      printf( "l = %d\n" , s->value );
      break;
      case FALSE :
      printf( " = %s\n" , s->bufptr );
      }
      */
    return s;
}


static void	itoxd_by_size( buf , num , size )
char	*buf;
ULONG	num;
opesize	size;
{
    switch( size ) {
    case LONGSIZE:
	itoxd( buf , num , 8 );
	break;
    case WORDSIZE:
	itoxd( buf , num , 4 );
	break;
    case BYTESIZE:
	itoxd( buf , num , 2 );
	break;
    default:	/* reduce warning message */
	break;
    }
}


static unsigned long	peek( adrs , size )
address	adrs;
opesize	size;
{
    unsigned long	rc = 0;
    
    switch( size ) {
    case BYTESIZE:
	rc = *(unsigned char*)( adrs + Ofst );
	break;
    case WORDSIZE:
	if( (long)adrs & 1 )
	    err( 0 , "Warning: peek.w ̈AhX(%x)ł\n" , adrs );
	rc = PEEK_WORD( adrs + Ofst );
	break;
    case LONGSIZE:
	if( (long)adrs & 1 )
	    err( 0 , "Warning: peek.l ̈AhX(%x)ł\n" , adrs );
	rc = PEEK_LONG( adrs + Ofst );
	break;
    default: /* reduce warning message */
	;
    }
    /*
      printf( "peek.%d(%x) = %x\n" , size , adrs , rc );
      */
    return rc;
}



static unsigned long	extend( arg , size )
unsigned long	arg;
opesize	size;
{
    unsigned long	rc = 0;    

    switch( size ) {
    case WORDSIZE:
	rc = (unsigned short)(signed char)arg;
	break;
    case LONGSIZE:
	rc = (unsigned long)(signed short)arg;
	break;
    default: /* reduce warning message */
	;
    }
    return rc;
}


static void	breakjob( value )
unsigned long	value;
{
    if( Eval_Count == 0 )			/* easy trick ? */
	Eval_Count = 1;
    Eval_ID = BREAKID;
    Eval_Break = value;
}


#ifdef SELF
main()
{
    char	Buffer[256];

    yydebug = 0;
    while(1) {
	gets( Buffer );
	Lexptr = Buffer;
	yyparse();
    }
}
#endif /* SELF */
