/*
 * PROJECT C Library, X68000 PROGRAMMING INTERFACE DEFINITION
 * --------------------------------------------------------------------
 * This file is written by the Project C Library Group,  and completely
 * in public domain. You can freely use, copy, modify, and redistribute
 * the whole contents, without this notice.
 * --------------------------------------------------------------------
 * $Id: getopt.c,v 1.1 1996/11/07 08:03:38 ryo freeze $
 */

/* System headers */
#include <stdio.h>
#include <stdlib.h>

char *optarg;													/* w|C^ */
int optind = -1;												/* ARGV ݂̌̃CfbNX */
int opterr = 1; 												/* G[\tO */
int _getopt_no_ordering;										/* ORDER tO */

/* File scope functions */
static void rotate_right (char *vector[], int size)
{
	char *temp;

	/* Ōۑ */
	temp = vector[size - 1];

	/* size [e[g */
	while (size--)
		vector[size] = vector[size - 1];

	/* 擪ɕۑĂ̂... */
	vector[0] = temp;
}

/* Functions */
int getopt (int argc, char *argv[], const char *options)
{
	int index;													/* ݂̃CfbNX */
	int next;													/* ̃CfbNX */
	int optchar;												/* IvV */
	char *string;												/* IvV */
	const char *ptr;											/* options Tp */

	static int init;											/* 񏉊Kv */
	static int savepoint;										/* ꊷp̋L|Cg */
	static int rotated; 										/* tO */
	static int comidx;											/* CfbNX */

#define ERR if (opterr) fprintf

	/* ̕KvΏ */
	if (init || optind < 0) {
		optind = 1;
		optarg = 0;
		rotated = 0;
		init = 0;
		comidx = 1;
		savepoint = 0;
	}

	/* {Jnʒuݒ */
	index = optind;

  again:

	/* o */
	string = argv[index];
	next = index + 1;

	/* łɏI肩H */
	if (string == 0) {
		if (savepoint > 0)
			optind = savepoint;
		init = 1;
		return EOF;
	}

	/* '-' Ŏn܂邩H */
	if (*string == '-') {

		/* tO|C^i߂ */
		string += comidx;

		/* m "-" HȂΕʂ̈ */
		if ((optchar = *string++) == '\0')
			goto normal_arg;

		/* ORDERING ̕KvΈ𕔕IɃ[e[g */
		if (savepoint > 0) {
			rotate_right (&argv[savepoint], index - savepoint + 1);
			rotated = 1;
			index = savepoint;
			savepoint++;
		}

		/* m "--" Ȃ΋Iɑ{Iɂ */
		if (optchar == '-' && *string == '\0' && comidx == 1){
			init = 1;
			optchar = EOF;
			goto goback;
		}

		/* IvVQ̒Ŷ邩ׂ */
		for (ptr = options; *ptr; ptr++)
			if (*ptr != ':' && *ptr == optchar)
				break;

		/* 𔺂ꍇȂ΃CfbNX͏ */
		if (*string == '\0' || ptr[1] == ':')
			comidx = 1;

		/* ȂΕIvVCfbNXZ */
		else {
			comidx++;
			index--;
		}

		/* ǌȂȂ... */
		if (*ptr == '\0') {
			ERR (stderr, "%s: unrecognized option '-%c'\n", argv[0], optchar);
			optchar = '?';
		}

		/* IvVw ':' Ȃ... */
		else if (ptr[1] == ':') {

			/*  argv Ɉ邩 */
			if (*string)
				optarg = string;

			/* optionsw肪 ?:: Ȃ΁A argv 炵Ȃ */
			else if (ptr[2] == ':')
				optarg = NULL;

			/* ̈ɂ邩 */
			else if (argv[next]) {

				/* ORDERING ̕KvΕIɓꊷ */
				if (rotated) {
					rotate_right (&argv[savepoint], next - savepoint + 1);
					index = savepoint;
				}

				/* Ȃ... */
				else
					index++;

				/* ̈Ԃ */
				optarg = argv[index];

			}

			/* Ȃ... */
			else {
				ERR (stderr, "%s: option '-%c' requires an argument\n", argv[0], optchar);
				optchar = '?';
			}

		}

	  goback:

		/* lݒ肵Ė߂ */
		rotated = 0;
		savepoint = 0;
		optind = index + 1;
		return optchar;

	}

	/* ʂ̈ */
	else {

	  normal_arg:

		/* ORDERING Kv邩H */
		if (_getopt_no_ordering) {
			init = 1;
			optind = index;
			return EOF;
		}

		/* ̈ʒuLÃIvV𒲂ׂ */
		else {
			if (savepoint == 0)
				savepoint = index;
			index++;
			goto again;
		}

	}
}

